#
#   Pyrex Top Level
#

import os
import sys
from time import time
import Version
from Scanning import PyrexScanner
import Errors
from Errors import PyrexError, CompileError, InternalError
import Parsing
import Code
from Pyrex.Utils import replace_suffix

verbose = 0

class CompilationOptions:
  #  show_version      boolean   Display version number
  #  use_listing_file  boolean   Generate a .lis file
  #  include_path      [string]  Directories to search for include files
  #  output_file       string    Name of generated .c file
  
  def __init__(self, defaults = None, **kw):
    self.include_path = []
    if defaults:
      self.__dict__.update(defaults.__dict__)
    self.__dict__.update(kw)

#
#  Set the default options depending on the platform
#

default_options = CompilationOptions(
  show_version = 0,
  use_listing_file = 0,
  output_file = None)
  
if sys.platform == "mac":
  from Pyrex.Mac.MacSystem import c_compile, c_link
  default_options.use_listing_file = 1


class CompilationResult:
  #  c_file           string or None
  #  h_file           string or None
  #  i_file           string or None
  #  listing_file     string or None
  #  object_file      string or None
  #  extension_file   string or None
  #  num_errors       integer
  #
  #  Used internally:
  #
  #  c_file_generated   boolean
  #  h_file_generated   boolean
  #  i_file_generated   boolean
  
  def __init__(self):
    self.c_file = None
    self.h_file = None
    self.listing_file = None
    self.object_file = None
    self.extension_file = None
    self.c_file_generated = 0
    self.h_file_generated = 0
    self.i_file_generated = 0


def extract_module_name(path):
  _, tail = os.path.split(path)
  name, _ = os.path.splitext(tail)
  return name

def parse(options, source):
  f = open(source, "r")
  s = PyrexScanner(f, source)
  s.include_directories = options.include_path
  try:
    tree = Parsing.p_module(s)
  finally:
    f.close()
  return tree

def compile(source, options = default_options,
    keep_files = 0,
    c_only = 0,
    obj_only = 0,
    dump_parse_tree = 0):
  try:
    result = CompilationResult()
    output_generated = 0
    cwd = os.getcwd()
    source = os.path.join(cwd, source)
    if options.use_listing_file:
      result.listing_file = replace_suffix(source, ".lis")
      Errors.open_listing_file(result.listing_file)
    try:
      import Scanning ###
      Scanning.get_lexicon() ### so as not to disturb parse timing
      if verbose:
        print "Parsing", source ###
      t0 = time() ###
      tree = parse(options, source)
      t1 = time() ###
      if verbose:
        print "Done (%.2f seconds)" % (t1 - t0) ###
      if dump_parse_tree:
        Parsing.print_parse_tree(sys.stdout, tree, 0)
      if options.output_file:
        result.c_file = os.path.join(cwd, options.output_file)
      else:
        result.c_file = replace_suffix(source, ".c")
      result.h_file = replace_suffix(result.c_file, ".h")
      result.i_file = replace_suffix(result.c_file, ".pxi")
      name = extract_module_name(source)
      if verbose:
        print "Processing parse tree" ###
      t0 = time() ###
      tree.process(name, result)
      t1 = time() ###
      if verbose:
        print "Done (%.2f seconds)" % (t1 - t0) ###
      output_generated = 1
    finally:
      Errors.close_listing_file()
  except CompileError:
    pass
  if not result.h_file_generated:
    result.h_file = None
  if not result.i_file_generated:
    result.i_file = None
  if not result.c_file_generated:
    result.c_file = None
  result.num_errors = Errors.num_errors
  if result.num_errors > 0 or not output_generated:
    return result
  if c_only:
    return result
  if verbose:
    print "Compiling", result.c_file ###
  result.object_file = c_compile(result.c_file)
  if obj_only:
    return result
  if verbose:
    print "Linking", result.object_file ###
  result.extension_file = c_link(result.object_file)
  return result

def main(command_line = 0, c_only = 0):
  args = sys.argv[1:]
  any_failures = 0
  if command_line:
    from CmdLine import parse_command_line
    options, sources = parse_command_line(args)
  else:
    options = default_options
    sources = args
  if options.show_version:
    print >>sys.stderr, "Pyrex version %s" % Version.version
  for source in sources:
    try:
      result = compile(source, options, c_only = c_only)
      if result.num_errors > 0:
        any_failures = 1
    except PyrexError, e:
      print >>sys.stderr, e
      any_failures = 1
  if any_failures:
    sys.exit(1)


